@@ -10,10 +10,10 @@
 #include <cstring>
 #include <sstream>
 #include "averr.h"
+#include <gbm.h>
 
 
 static const AVRational US_RATIONAL{1,1000000} ;
-#define AUDIO_RATE 44100
 
 // av_register_all was deprecated in 58.9.100, removed in 59.0.100
 #if LIBAVCODEC_VERSION_INT < AV_VERSION_INT(59, 0, 100)
@@ -45,19 +45,35 @@ void FrameWriter::init_hw_accel()
 
 void FrameWriter::load_codec_options(AVDictionary **dict)
 {
-    static const std::map<std::string, std::string> default_x264_options = {
-         {"tune", "zerolatency"},
-         {"preset", "ultrafast"},
-         {"crf", "20"},
+    using CodecOptions = std::map<std::string, std::string>;
+
+    static const CodecOptions default_x264_options = {
+        {"tune", "zerolatency"},
+        {"preset", "ultrafast"},
+        {"crf", "20"},
     };
 
-    if (params.codec.find("libx264") != std::string::npos ||
-        params.codec.find("libx265") != std::string::npos)
+    static const CodecOptions default_libvpx_options = {
+        {"cpu-used", "5"},
+        {"deadline", "realtime"},
+    };
+
+    static const std::map<std::string, const CodecOptions&> default_codec_options = {
+        {"libx264", default_x264_options},
+        {"libx265", default_x264_options},
+        {"libvpx", default_libvpx_options},
+    };
+
+    for (const auto& opts : default_codec_options)
     {
-        for (const auto& param : default_x264_options)
+        if (params.codec.find(opts.first) != std::string::npos)
         {
-            if (!params.codec_options.count(param.first))
-                params.codec_options[param.first] = param.second;
+            for (const auto& param : opts.second)
+            {
+                if (!params.codec_options.count(param.first))
+                    params.codec_options[param.first] = param.second;
+            }
+            break;
         }
     }
 
@@ -68,6 +84,15 @@ void FrameWriter::load_codec_options(AVDictionary **dict)
     }
 }
 
+void FrameWriter::load_audio_codec_options(AVDictionary **dict)
+{
+    for (auto& opt : params.audio_codec_options)
+    {
+        std::cout << "Setting codec option: " << opt.first << "=" << opt.second << std::endl;
+        av_dict_set(dict, opt.first.c_str(), opt.second.c_str(), 0);
+    }
+}
+
 bool is_fmt_supported(AVPixelFormat fmt, const AVPixelFormat *supported)
 {
     for (int i = 0; supported[i] != AV_PIX_FMT_NONE; i++)
@@ -100,12 +125,48 @@ AVPixelFormat FrameWriter::get_input_format()
     case INPUT_FORMAT_X2BGR10:
         return AV_PIX_FMT_X2BGR10LE;
 #endif
+    case INPUT_FORMAT_RGBX64:
+        return AV_PIX_FMT_RGBA64LE;
+    case INPUT_FORMAT_BGRX64:
+        return AV_PIX_FMT_BGRA64LE;
+#if LIBAVUTIL_VERSION_INT >= AV_VERSION_INT(57, 33, 101)
+    case INPUT_FORMAT_RGBX64F:
+        return AV_PIX_FMT_RGBAF16LE;
+#endif
+    case INPUT_FORMAT_DMABUF:
+        return AV_PIX_FMT_VAAPI;
     default:
         std::cerr << "Unknown format: " << params.format << std::endl;
         std::exit(-1);
     }
 }
 
+static const struct {
+    int drm;
+    AVPixelFormat av;
+} drm_av_format_table [] = {
+    { GBM_FORMAT_ARGB8888, AV_PIX_FMT_BGRA },
+    { GBM_FORMAT_XRGB8888, AV_PIX_FMT_BGR0 },
+    { GBM_FORMAT_ABGR8888, AV_PIX_FMT_RGBA },
+    { GBM_FORMAT_XBGR8888, AV_PIX_FMT_RGB0 },
+    { GBM_FORMAT_RGBA8888, AV_PIX_FMT_ABGR },
+    { GBM_FORMAT_RGBX8888, AV_PIX_FMT_0BGR },
+    { GBM_FORMAT_BGRA8888, AV_PIX_FMT_ARGB },
+    { GBM_FORMAT_BGRX8888, AV_PIX_FMT_0RGB },
+    { GBM_FORMAT_XRGB2101010, AV_PIX_FMT_X2RGB10 },
+};
+
+static AVPixelFormat get_drm_av_format(int fmt)
+{
+    for (size_t i = 0; i < sizeof(drm_av_format_table) / sizeof(drm_av_format_table[0]); ++i) {
+        if (drm_av_format_table[i].drm == fmt) {
+            return drm_av_format_table[i].av;
+        }
+    }
+    std::cerr << "Failed to find AV format for" << fmt;
+    return AV_PIX_FMT_RGBA;
+}
+
 AVPixelFormat FrameWriter::lookup_pixel_format(std::string pix_fmt)
 {
     AVPixelFormat fmt = av_get_pix_fmt(pix_fmt.c_str());
@@ -117,13 +178,15 @@ AVPixelFormat FrameWriter::lookup_pixel_format(std::string pix_fmt)
     std::exit(-1);
 }
 
-AVPixelFormat FrameWriter::choose_sw_format(const AVCodec *codec)
+AVPixelFormat FrameWriter::handle_buffersink_pix_fmt(const AVCodec *codec)
 {
-    auto in_fmt = get_input_format();
 
+    // Return with user chosen format
     if (!params.pix_fmt.empty())
         return lookup_pixel_format(params.pix_fmt);
 
+    auto in_fmt = get_input_format();
+
     /* For codecs such as rawvideo no supported formats are listed */
     if (!codec->pix_fmts)
         return in_fmt;
@@ -133,22 +196,26 @@ AVPixelFormat FrameWriter::choose_sw_format(const AVCodec *codec)
     if (is_fmt_supported(in_fmt, codec->pix_fmts))
         return in_fmt;
 
-    /* Otherwise, try to use the already tested YUV420p */
-    if (is_fmt_supported(AV_PIX_FMT_YUV420P, codec->pix_fmts))
-        return AV_PIX_FMT_YUV420P;
-
-    /* Lastly, use the first supported format */
-    return codec->pix_fmts[0];
+    /* Choose the format supported by the codec which best approximates the
+     * input fmt. */
+    AVPixelFormat best_format = AV_PIX_FMT_NONE;
+    for (int i = 0; codec->pix_fmts[i] != AV_PIX_FMT_NONE; i++) {
+        int loss = 0;
+        best_format = av_find_best_pix_fmt_of_2(best_format,
+            codec->pix_fmts[i], in_fmt, false, &loss);
+    }
+    return best_format;
 }
 
 void FrameWriter::init_video_filters(const AVCodec *codec)
 {
-    if (params.codec.find("vaapi") != std::string::npos) {
-        if (params.video_filter == "null") {
-            // Add `hwupload,scale_vaapi=format=nv12` by default
-            // It is necessary for conversion to a proper format.
-            params.video_filter = "hwupload,scale_vaapi=format=nv12";
+    if (params.framerate != 0){
+        if (params.video_filter != "null" && params.video_filter.find("fps") == std::string::npos) {
+            params.video_filter += ",fps=" + std::to_string(params.framerate);
         }
+        else if (params.video_filter == "null"){
+            params.video_filter = "fps=" + std::to_string(params.framerate);
+        } 
     }
 
     this->videoFilterGraph = avfilter_graph_alloc();
@@ -162,6 +229,32 @@ void FrameWriter::init_video_filters(const AVCodec *codec)
         exit(-1);
     }
 
+    if (this->hw_device_context) {
+        this->hw_frame_context = av_hwframe_ctx_alloc(this->hw_device_context);
+        AVHWFramesContext *hwfc = reinterpret_cast<AVHWFramesContext*>(this->hw_frame_context->data);
+        hwfc->format = AV_PIX_FMT_VAAPI;
+        hwfc->sw_format = AV_PIX_FMT_NV12;
+        hwfc->width = params.width;
+        hwfc->height = params.height;
+        int err = av_hwframe_ctx_init(this->hw_frame_context);
+        if (err < 0) {
+            std::cerr << "Cannot create hw frames context: " << averr(err) << std::endl;
+            exit(-1);
+        }
+
+        this->hw_frame_context_in = av_hwframe_ctx_alloc(this->hw_device_context);
+        hwfc = reinterpret_cast<AVHWFramesContext*>(this->hw_frame_context_in->data);
+        hwfc->format = AV_PIX_FMT_VAAPI;
+        hwfc->sw_format = get_drm_av_format(params.drm_format);
+        hwfc->width = params.width;
+        hwfc->height = params.height;
+        err = av_hwframe_ctx_init(this->hw_frame_context_in);
+        if (err < 0) {
+            std::cerr << "Cannot create hw frames context: " << averr(err) << std::endl;
+            exit(-1);
+        }
+    }
+
     // Build the configuration of the 'buffer' filter.
     // See: ffmpeg -h filter=buffer
     // See: https://ffmpeg.org/ffmpeg-filters.html#buffer
@@ -169,6 +262,9 @@ void FrameWriter::init_video_filters(const AVCodec *codec)
     buffer_filter_config << "video_size=" << params.width << "x" << params.height;
     buffer_filter_config << ":pix_fmt=" << (int)this->get_input_format();
     buffer_filter_config << ":time_base=" << US_RATIONAL.num << "/" << US_RATIONAL.den;
+    if (params.buffrate != 0) {
+        buffer_filter_config << ":frame_rate=" << params.buffrate;
+    }
     buffer_filter_config << ":pixel_aspect=1/1";
 
     int err = avfilter_graph_create_filter(&this->videoFilterSourceCtx, source,
@@ -178,6 +274,17 @@ void FrameWriter::init_video_filters(const AVCodec *codec)
         exit(-1);
     }
 
+    AVBufferSrcParameters *p = av_buffersrc_parameters_alloc();
+    memset(p, 0, sizeof(*p));
+    p->format = AV_PIX_FMT_NONE;
+    p->hw_frames_ctx = this->hw_frame_context_in;
+    err = av_buffersrc_parameters_set(this->videoFilterSourceCtx, p);
+    av_free(p);
+    if (err < 0) {
+         std::cerr << "Cannot set hwcontext filter in: " << averr(err) << std::endl;;
+         exit(-1);
+    }
+
     err = avfilter_graph_create_filter(&this->videoFilterSinkCtx, sink, "Sink",
         NULL, NULL, this->videoFilterGraph);
     if (err < 0) {
@@ -188,26 +295,14 @@ void FrameWriter::init_video_filters(const AVCodec *codec)
     // We also need to tell the sink which pixel formats are supported.
     // by the video encoder. codevIndicate to our sink  pixel formats
     // are accepted by our codec.
-    const AVPixelFormat *supported_pix_fmts = codec->pix_fmts;
-    static const AVPixelFormat only_yuv420p[] =
+    const AVPixelFormat picked_pix_fmt[] =
     {
-        AV_PIX_FMT_YUV420P,
+        handle_buffersink_pix_fmt(codec),
         AV_PIX_FMT_NONE
     };
 
-    // Force pixel format to yuv420p if requested and possible
-    if (params.force_yuv) {
-        if (!supported_pix_fmts ||
-            is_fmt_supported(AV_PIX_FMT_YUV420P, supported_pix_fmts)) {
-            supported_pix_fmts = only_yuv420p ;
-        } else {
-            std::cerr << "Ignoring request to force yuv420p, " <<
-                "because it is not supported by the codec\n";
-        }
-    }
-
     err = av_opt_set_int_list(this->videoFilterSinkCtx, "pix_fmts",
-        supported_pix_fmts, AV_PIX_FMT_NONE, AV_OPT_SEARCH_CHILDREN);
+        picked_pix_fmt, AV_PIX_FMT_NONE, AV_OPT_SEARCH_CHILDREN);
 
     if (err < 0) {
         std::cerr << "Failed to set pix_fmts: " << averr(err) << std::endl;;
@@ -286,8 +381,6 @@ void FrameWriter::init_video_filters(const AVCodec *codec)
     this->videoCodecCtx->framerate = filter_output->frame_rate; // can be 1/0 if unknown
     this->videoCodecCtx->sample_aspect_ratio = filter_output->sample_aspect_ratio;
 
-    this->hw_frame_context = av_buffersink_get_hw_frames_ctx(
-        this->videoFilterSinkCtx);
     avfilter_inout_free(&inputs);
     avfilter_inout_free(&outputs);
 }
@@ -316,6 +409,10 @@ void FrameWriter::init_video_stream()
     videoCodecCtx->height     = params.height;
     videoCodecCtx->time_base  = US_RATIONAL;
     videoCodecCtx->color_range = AVCOL_RANGE_JPEG;
+    if (params.framerate) {
+    std::cout << "Framerate: " << params.framerate << std::endl;
+    }
+
     if (params.bframes != -1)
         videoCodecCtx->max_b_frames = params.bframes;
 
@@ -370,11 +467,11 @@ static uint64_t get_codec_channel_layout(const AVCodec *codec)
       return codec->channel_layouts[0];
 }
 
-static enum AVSampleFormat get_codec_sample_fmt(const AVCodec *codec)
+static enum AVSampleFormat get_codec_auto_sample_fmt(const AVCodec *codec)
 {
     int i = 0;
     if (!codec->sample_fmts)
-        return AV_SAMPLE_FMT_S16;
+        return av_get_sample_fmt(FALLBACK_AUDIO_SAMPLE_FMT);
     while (1) {
         if (codec->sample_fmts[i] == -1)
             break;
@@ -385,12 +482,44 @@ static enum AVSampleFormat get_codec_sample_fmt(const AVCodec *codec)
     return codec->sample_fmts[0];
 }
 
+bool check_fmt_available(const AVCodec *codec, AVSampleFormat fmt){
+    for (const enum AVSampleFormat *sample_ptr = codec -> sample_fmts; *sample_ptr != -1; sample_ptr++)
+    {
+        if (*sample_ptr == fmt)
+        {
+            return true;
+        }
+    }
+    return false;
+}
+
+static enum AVSampleFormat convert_codec_sample_fmt(const AVCodec *codec, std::string requested_fmt)
+{
+    static enum AVSampleFormat converted_fmt = av_get_sample_fmt(requested_fmt.c_str());
+    if (converted_fmt == AV_SAMPLE_FMT_NONE)
+    {
+	std::cout << "Failed to find the given sample format: " << requested_fmt << std::endl;
+	std::exit(-1);
+    } else if (!codec->sample_fmts || check_fmt_available(codec, converted_fmt))
+    {
+        std::cout << "Using sample format " << av_get_sample_fmt_name(converted_fmt) << " for audio codec " << codec->name << std::endl;
+        return converted_fmt;
+    } else
+    {
+	std::cout << "Codec " << codec->name << " does not support sample format " << av_get_sample_fmt_name(converted_fmt) << std::endl;
+	std::exit(-1);
+    }
+}
+
 void FrameWriter::init_audio_stream()
 {
-    const AVCodec* codec = avcodec_find_encoder_by_name("aac");
+    AVDictionary *options = NULL;
+    load_codec_options(&options);
+    
+    const AVCodec* codec = avcodec_find_encoder_by_name(params.audio_codec.c_str());
     if (!codec)
     {
-        std::cerr << "Failed to find the aac codec" << std::endl;
+        std::cerr << "Failed to find the given audio codec: " << params.audio_codec << std::endl;
         std::exit(-1);
     }
 
@@ -402,10 +531,16 @@ void FrameWriter::init_audio_stream()
     }
 
     audioCodecCtx = avcodec_alloc_context3(codec);
-    audioCodecCtx->bit_rate = lrintf(128000.0f);
-    audioCodecCtx->sample_fmt = get_codec_sample_fmt(codec);
+    if (params.sample_fmt.size() == 0) 
+    {
+        audioCodecCtx->sample_fmt = get_codec_auto_sample_fmt(codec);
+        std::cout << "Choosing sample format " << av_get_sample_fmt_name(audioCodecCtx->sample_fmt) << " for audio codec " << codec->name << std::endl;
+    } else 
+    {
+        audioCodecCtx->sample_fmt = convert_codec_sample_fmt(codec, params.sample_fmt);
+    }
     audioCodecCtx->channel_layout = get_codec_channel_layout(codec);
-    audioCodecCtx->sample_rate = AUDIO_RATE;
+    audioCodecCtx->sample_rate = params.sample_rate;
     audioCodecCtx->time_base = (AVRational) { 1, 1000 };
     audioCodecCtx->channels = av_get_channel_layout_nb_channels(audioCodecCtx->channel_layout);
 
@@ -426,7 +561,7 @@ void FrameWriter::init_audio_stream()
         std::exit(-1);
     }
 
-    av_opt_set_int(swrCtx, "in_sample_rate", AUDIO_RATE, 0);
+    av_opt_set_int(swrCtx, "in_sample_rate", params.sample_rate, 0);
     av_opt_set_int(swrCtx, "out_sample_rate", audioCodecCtx->sample_rate, 0);
     av_opt_set_sample_fmt(swrCtx, "in_sample_fmt", AV_SAMPLE_FMT_FLT, 0);
     av_opt_set_sample_fmt(swrCtx, "out_sample_fmt", audioCodecCtx->sample_fmt, 0);
@@ -541,35 +676,8 @@ void FrameWriter::encode(AVCodecContext *enc_ctx, AVFrame *frame, AVPacket *pkt)
     }
 }
 
-AVFrame *FrameWriter::prepare_frame_data(const uint8_t* pixels, bool y_invert)
+bool FrameWriter::push_frame(AVFrame *frame, int64_t usec)
 {
-    /* Calculate data after y-inversion */
-    int stride[] = {int(params.stride)};
-    const uint8_t *formatted_pixels = pixels;
-    if (y_invert)
-    {
-        formatted_pixels += stride[0] * (params.height - 1);
-        stride[0] *= -1;
-    }
-
-    auto frame = av_frame_alloc();
-    if (!frame) {
-        std::cerr << "Failed to allocate frame!" << std::endl;
-        return NULL;
-    }
-
-    frame->data[0] = (uint8_t*)formatted_pixels;
-    frame->linesize[0] = stride[0];
-    frame->format = get_input_format();
-    frame->width = params.width;
-    frame->height = params.height;
-
-    return frame;
-}
-
-bool FrameWriter::add_frame(const uint8_t* pixels, int64_t usec, bool y_invert)
-{
-    auto frame = prepare_frame_data(pixels, y_invert);
     frame->pts = usec; // We use time_base = 1/US_RATE
 
     // Push the RGB frame into the filtergraph */
@@ -592,6 +700,7 @@ bool FrameWriter::add_frame(const uint8_t* pixels, int64_t usec, bool y_invert)
         if (err == AVERROR(EAGAIN)) {
             // Not an error. No frame available.
             // Try again later.
+            av_frame_free(&filtered_frame);
             break;
         } else if (err == AVERROR_EOF) {
             // There will be no more output frames on this sink.
@@ -606,26 +715,116 @@ bool FrameWriter::add_frame(const uint8_t* pixels, int64_t usec, bool y_invert)
         filtered_frame->pict_type = AV_PICTURE_TYPE_NONE;
 
         // So we have a frame. Encode it!
-        AVPacket pkt;
-        av_init_packet(&pkt);
-        pkt.data = NULL;
-        pkt.size = 0;
+        AVPacket *pkt = av_packet_alloc();
+        pkt->data = NULL;
+        pkt->size = 0;
 
-        encode(videoCodecCtx, filtered_frame, &pkt);
+        encode(videoCodecCtx, filtered_frame, pkt);
         av_frame_free(&filtered_frame);
+        av_packet_free(&pkt);
     }
 
     av_frame_free(&frame);
     return true;
 }
 
+bool FrameWriter::add_frame(const uint8_t* pixels, int64_t usec, bool y_invert)
+{
+    /* Calculate data after y-inversion */
+    int stride[] = {int(params.stride)};
+    const uint8_t *formatted_pixels = pixels;
+    if (y_invert)
+    {
+        formatted_pixels += stride[0] * (params.height - 1);
+        stride[0] *= -1;
+    }
+
+    auto frame = av_frame_alloc();
+    if (!frame) {
+        std::cerr << "Failed to allocate frame!" << std::endl;
+        return false;
+    }
+
+    frame->data[0] = (uint8_t*)formatted_pixels;
+    frame->linesize[0] = stride[0];
+    frame->format = get_input_format();
+    frame->width = params.width;
+    frame->height = params.height;
+
+    return push_frame(frame, usec);
+}
+
+bool FrameWriter::add_frame(struct gbm_bo *bo, int64_t usec, bool y_invert)
+{
+    if (y_invert)
+    {
+        std::cerr << "Y_INVERT not supported with dmabuf" << std::endl;
+        return false;
+    }
+
+    auto frame = av_frame_alloc();
+    if (!frame)
+    {
+        std::cerr << "Failed to allocate frame!" << std::endl;
+        return false;
+    }
+
+    if (mapped_frames.find(bo) == mapped_frames.end()) {
+        auto vaapi_frame = av_frame_alloc();
+        if (!vaapi_frame) {
+            std::cerr << "Failed to allocate frame!" << std::endl;
+            return false;
+        }
+
+        AVDRMFrameDescriptor *desc = (AVDRMFrameDescriptor*) av_mallocz(sizeof(AVDRMFrameDescriptor));
+        desc->nb_layers = 1;
+        desc->nb_objects = 1;
+        desc->objects[0].fd = gbm_bo_get_fd(bo);
+        desc->objects[0].format_modifier = gbm_bo_get_modifier(bo);
+        desc->objects[0].size = gbm_bo_get_stride(bo) * gbm_bo_get_height(bo);
+        desc->layers[0].format = gbm_bo_get_format(bo);
+        desc->layers[0].nb_planes = gbm_bo_get_plane_count(bo);
+        for (int i = 0; i < gbm_bo_get_plane_count(bo); ++i) {
+            desc->layers[0].planes[i].object_index = 0;
+            desc->layers[0].planes[i].pitch = gbm_bo_get_stride_for_plane(bo, i);
+            desc->layers[0].planes[i].offset = gbm_bo_get_offset(bo, i);
+        }
+
+        frame->width = gbm_bo_get_width(bo);
+        frame->height = gbm_bo_get_height(bo);
+        frame->format = AV_PIX_FMT_DRM_PRIME;
+        frame->data[0] = reinterpret_cast<uint8_t*>(desc);
+        frame->buf[0] = av_buffer_create(frame->data[0], sizeof(*desc),
+            [](void *, uint8_t *data) {
+                av_free(data);
+        }, frame, 0);
+
+        vaapi_frame->format = AV_PIX_FMT_VAAPI;
+        vaapi_frame->hw_frames_ctx = av_buffer_ref(this->hw_frame_context_in);
+
+        int ret = av_hwframe_map(vaapi_frame, frame, AV_HWFRAME_MAP_READ);
+        av_frame_unref(frame);
+        if (ret < 0)
+        {
+            std::cerr << "Failed to map vaapi frame " << averr(ret) << std::endl;
+            return false;
+        }
+
+        mapped_frames[bo] = vaapi_frame;
+    }
+
+    av_frame_ref(frame, mapped_frames[bo]);
+    return push_frame(frame, usec);
+}
+
 #ifdef HAVE_PULSE
 #define SRC_RATE 1e6
 #define DST_RATE 1e3
 
-static int64_t conv_audio_pts(SwrContext *ctx, int64_t in)
+static int64_t conv_audio_pts(SwrContext *ctx, int64_t in, int sample_rate)
 {
-    int64_t d = (int64_t) AUDIO_RATE * AUDIO_RATE;
+    //int64_t d = (int64_t) AUDIO_RATE * AUDIO_RATE;
+    int64_t d = (int64_t) sample_rate * sample_rate;
 
     /* Convert from audio_src_tb to 1/(src_samplerate * dst_samplerate) */
     in = av_rescale_rnd(in, d, SRC_RATE, AV_ROUND_NEAR_INF);
@@ -639,12 +838,12 @@ static int64_t conv_audio_pts(SwrContext *ctx, int64_t in)
 
 void FrameWriter::send_audio_pkt(AVFrame *frame)
 {
-    AVPacket pkt;
-    av_init_packet(&pkt);
-    pkt.data = NULL;
-    pkt.size = 0;
+    AVPacket *pkt = av_packet_alloc();
+    pkt->data = NULL;
+    pkt->size = 0;
 
-    encode(audioCodecCtx, frame, &pkt);
+    encode(audioCodecCtx, frame, pkt);
+    av_packet_free(&pkt);
 }
 
 size_t FrameWriter::get_audio_buffer_size()
@@ -655,7 +854,7 @@ size_t FrameWriter::get_audio_buffer_size()
 void FrameWriter::add_audio(const void* buffer)
 {
     AVFrame *inputf = av_frame_alloc();
-    inputf->sample_rate    = AUDIO_RATE;
+    inputf->sample_rate    = params.sample_rate;
     inputf->format         = AV_SAMPLE_FMT_FLT;
     inputf->channel_layout = AV_CH_LAYOUT_STEREO;
     inputf->nb_samples     = audioCodecCtx->frame_size;
@@ -670,7 +869,7 @@ void FrameWriter::add_audio(const void* buffer)
     outputf->nb_samples     = audioCodecCtx->frame_size;
     av_frame_get_buffer(outputf, 0);
 
-    outputf->pts = conv_audio_pts(swrCtx, INT64_MIN);
+    outputf->pts = conv_audio_pts(swrCtx, INT64_MIN, params.sample_rate);
     swr_convert_frame(swrCtx, outputf, inputf);
 
     send_audio_pkt(outputf);
@@ -719,14 +918,13 @@ void FrameWriter::finish_frame(AVCodecContext *enc_ctx, AVPacket& pkt)
 FrameWriter::~FrameWriter()
 {
     // Writing the delayed frames:
-    AVPacket pkt;
-    av_init_packet(&pkt);
+    AVPacket *pkt = av_packet_alloc();
 
-    encode(videoCodecCtx, NULL, &pkt);
+    encode(videoCodecCtx, NULL, pkt);
 #ifdef HAVE_PULSE
     if (params.enable_audio)
     {
-        encode(audioCodecCtx, NULL, &pkt);
+        encode(audioCodecCtx, NULL, pkt);
     }
 #endif
     // Writing the end of the file.
@@ -736,12 +934,13 @@ FrameWriter::~FrameWriter()
     if (outputFmt && (!(outputFmt->flags & AVFMT_NOFILE)))
         avio_closep(&fmtCtx->pb);
 
-    avcodec_free_context(&videoCodecCtx);
     // Freeing all the allocated memory:
+    avcodec_free_context(&videoCodecCtx);
 #ifdef HAVE_PULSE
     if (params.enable_audio)
         avcodec_free_context(&audioCodecCtx);
 #endif
+    av_packet_free(&pkt);
     // TODO: free all the hw accel
     avformat_free_context(fmtCtx);
 }
