@@ -18,10 +18,15 @@
 #include <signal.h>
 #include <unistd.h>
 #include <wayland-client-protocol.h>
+#include <gbm.h>
+#include <fcntl.h>
 
 #include "frame-writer.hpp"
+#include "buffer-pool.hpp"
 #include "wlr-screencopy-unstable-v1-client-protocol.h"
 #include "xdg-output-unstable-v1-client-protocol.h"
+#include "linux-dmabuf-unstable-v1-client-protocol.h"
+#include "wl-drm-client-protocol.h"
 
 #include "config.h"
 
@@ -32,12 +37,20 @@ PulseReaderParams pulseParams;
 
 #define MAX_FRAME_FAILURES 16
 
+static const int GRACEFUL_TERMINATION_SIGNALS[] = { SIGTERM, SIGINT, SIGHUP };
+
 std::mutex frame_writer_mutex, frame_writer_pending_mutex;
 std::unique_ptr<FrameWriter> frame_writer;
 
+static int drm_fd = -1;
+static struct gbm_device *gbm_device = NULL;
+static std::string drm_device_name;
+
 static struct wl_shm *shm = NULL;
 static struct zxdg_output_manager_v1 *xdg_output_manager = NULL;
 static struct zwlr_screencopy_manager_v1 *screencopy_manager = NULL;
+static struct zwp_linux_dmabuf_v1 *dmabuf = NULL;
+static struct wl_drm *drm = NULL;
 void request_next_frame();
 
 struct wf_recorder_output
@@ -107,26 +120,23 @@ const zxdg_output_v1_listener xdg_output_implementation = {
     .description = handle_xdg_output_description
 };
 
-struct wf_buffer
+struct wf_buffer : public buffer_pool_buf
 {
-    struct wl_buffer *wl_buffer;
-    void *data;
+    struct gbm_bo *bo = nullptr;
+    struct wl_buffer *wl_buffer = nullptr;
+    void *data = nullptr;
     enum wl_shm_format format;
+    int drm_format;
     int width, height, stride;
     bool y_invert;
 
     timespec presented;
     uint64_t base_usec;
-
-    std::atomic<bool> released{true}; // if the buffer can be used to store new pending frames
-    std::atomic<bool> available{false}; // if the buffer can be used to feed the encoder
 };
 
 std::atomic<bool> exit_main_loop{false};
 
-#define MAX_BUFFERS 16
-wf_buffer buffers[MAX_BUFFERS];
-size_t active_buffer = 0;
+buffer_pool<wf_buffer, 16> buffers;
 
 bool buffer_copy_done = false;
 
@@ -180,11 +190,17 @@ static struct wl_buffer *create_shm_buffer(uint32_t fmt,
 }
 
 static bool use_damage = true;
+static bool use_dmabuf = false;
+static bool use_hwupload = false;
 
 static void frame_handle_buffer(void *, struct zwlr_screencopy_frame_v1 *frame, uint32_t format,
     uint32_t width, uint32_t height, uint32_t stride)
 {
-    auto& buffer = buffers[active_buffer];
+    if (use_dmabuf) {
+        return;
+    }
+
+    auto& buffer = buffers.capture();
 
     buffer.format = (wl_shm_format)format;
     buffer.width = width;
@@ -215,7 +231,7 @@ static void frame_handle_buffer(void *, struct zwlr_screencopy_frame_v1 *frame,
 }
 
 static void frame_handle_flags(void*, struct zwlr_screencopy_frame_v1 *, uint32_t flags) {
-    buffers[active_buffer].y_invert = flags & ZWLR_SCREENCOPY_FRAME_V1_FLAGS_Y_INVERT;
+    buffers.capture().y_invert = flags & ZWLR_SCREENCOPY_FRAME_V1_FLAGS_Y_INVERT;
 }
 
 int32_t frame_failed_cnt = 0;
@@ -223,7 +239,7 @@ int32_t frame_failed_cnt = 0;
 static void frame_handle_ready(void *, struct zwlr_screencopy_frame_v1 *,
     uint32_t tv_sec_hi, uint32_t tv_sec_low, uint32_t tv_nsec) {
 
-    auto& buffer = buffers[active_buffer];
+    auto& buffer = buffers.capture();
     buffer_copy_done = true;
     buffer.presented.tv_sec = ((1ll * tv_sec_hi) << 32ll) | tv_sec_low;
     buffer.presented.tv_nsec = tv_nsec;
@@ -233,6 +249,7 @@ static void frame_handle_ready(void *, struct zwlr_screencopy_frame_v1 *,
 static void frame_handle_failed(void *, struct zwlr_screencopy_frame_v1 *) {
     std::cerr << "Failed to copy frame, retrying..." << std::endl;
     ++frame_failed_cnt;
+    request_next_frame();
     if (frame_failed_cnt > MAX_FRAME_FAILURES)
     {
         std::cerr << "Failed to copy frame too many times, exiting!" << std::endl;
@@ -245,12 +262,128 @@ static void frame_handle_damage(void *, struct zwlr_screencopy_frame_v1 *,
 {
 }
 
+static void dmabuf_created(void *data, struct zwp_linux_buffer_params_v1 *,
+    struct wl_buffer *wl_buffer) {
+
+    auto& buffer = buffers.capture();
+    buffer.wl_buffer = wl_buffer;
+
+    zwlr_screencopy_frame_v1 *frame = (zwlr_screencopy_frame_v1*) data;
+
+    if (use_damage) {
+        zwlr_screencopy_frame_v1_copy_with_damage(frame, buffer.wl_buffer);
+    } else {
+        zwlr_screencopy_frame_v1_copy(frame, buffer.wl_buffer);
+    }
+}
+
+static void dmabuf_failed(void *, struct zwp_linux_buffer_params_v1 *) {
+    std::cerr << "Failed to create dmabuf" << std::endl;
+    exit_main_loop = true;
+}
+
+static const struct zwp_linux_buffer_params_v1_listener params_listener = {
+    .created = dmabuf_created,
+    .failed = dmabuf_failed,
+};
+
+static wl_shm_format drm_to_wl_shm_format(uint32_t format)
+{
+    if (format == GBM_FORMAT_ARGB8888) {
+        return WL_SHM_FORMAT_ARGB8888;
+    } else if (format == GBM_FORMAT_XRGB8888) {
+        return WL_SHM_FORMAT_XRGB8888;
+    } else {
+        return (wl_shm_format)format;
+    }
+}
+
+static void frame_handle_linux_dmabuf(void *, struct zwlr_screencopy_frame_v1 *frame,
+    uint32_t format, uint32_t width, uint32_t height)
+{
+    if (!use_dmabuf) {
+        return;
+    }
+
+    auto& buffer = buffers.capture();
+
+    buffer.format = drm_to_wl_shm_format(format);
+    buffer.drm_format = format;
+    buffer.width = width;
+    buffer.height = height;
+
+    if (!buffer.wl_buffer) {
+        const uint64_t modifier = 0; // DRM_FORMAT_MOD_LINEAR
+        buffer.bo = gbm_bo_create_with_modifiers(gbm_device, buffer.width,
+            buffer.height, format, &modifier, 1);
+        if (buffer.bo == NULL)
+        {
+            buffer.bo = gbm_bo_create(gbm_device, buffer.width,
+                buffer.height, format, GBM_BO_USE_LINEAR | GBM_BO_USE_RENDERING);
+        }
+        if (buffer.bo == NULL)
+        {
+            std::cerr << "Failed to create gbm bo" << std::endl;
+            exit_main_loop = true;
+            return;
+        }
+
+        buffer.stride = gbm_bo_get_stride(buffer.bo);
+
+        struct zwp_linux_buffer_params_v1 *params =
+            zwp_linux_dmabuf_v1_create_params(dmabuf);
+
+        uint64_t mod = gbm_bo_get_modifier(buffer.bo);
+        zwp_linux_buffer_params_v1_add(params,
+            gbm_bo_get_fd(buffer.bo), 0,
+            gbm_bo_get_offset(buffer.bo, 0),
+            gbm_bo_get_stride(buffer.bo),
+            mod >> 32, mod & 0xffffffff);
+
+        zwp_linux_buffer_params_v1_add_listener(params, &params_listener, frame);
+
+        zwp_linux_buffer_params_v1_create(params, buffer.width,
+            buffer.height, format, 0);
+    } else {
+        if (use_damage) {
+            zwlr_screencopy_frame_v1_copy_with_damage(frame, buffer.wl_buffer);
+        } else {
+            zwlr_screencopy_frame_v1_copy(frame, buffer.wl_buffer);
+        }
+    }
+}
+
+static void frame_handle_buffer_done(void *, struct zwlr_screencopy_frame_v1 *) {
+}
+
 static const struct zwlr_screencopy_frame_v1_listener frame_listener = {
     .buffer = frame_handle_buffer,
     .flags = frame_handle_flags,
     .ready = frame_handle_ready,
     .failed = frame_handle_failed,
     .damage = frame_handle_damage,
+    .linux_dmabuf = frame_handle_linux_dmabuf,
+    .buffer_done = frame_handle_buffer_done,
+};
+
+static void drm_handle_device(void *, struct wl_drm *, const char *name) {
+    drm_device_name = name;
+}
+
+static void drm_handle_format(void *, struct wl_drm *, uint32_t) {
+}
+
+static void drm_handle_authenticated(void *, struct wl_drm *) {
+}
+
+static void drm_handle_capabilities(void *, struct wl_drm *, uint32_t) {
+}
+
+static const struct wl_drm_listener drm_listener = {
+    .device = drm_handle_device,
+    .format = drm_handle_format,
+    .authenticated = drm_handle_authenticated,
+    .capabilities = drm_handle_capabilities,
 };
 
 static void handle_global(void*, struct wl_registry *registry,
@@ -270,13 +403,23 @@ static void handle_global(void*, struct wl_registry *registry,
     else if (strcmp(interface, zwlr_screencopy_manager_v1_interface.name) == 0)
     {
         screencopy_manager = (zwlr_screencopy_manager_v1*) wl_registry_bind(registry, name,
-            &zwlr_screencopy_manager_v1_interface, 2);
+            &zwlr_screencopy_manager_v1_interface, 3);
     }
     else if (strcmp(interface, zxdg_output_manager_v1_interface.name) == 0)
     {
         xdg_output_manager = (zxdg_output_manager_v1*) wl_registry_bind(registry, name,
             &zxdg_output_manager_v1_interface, 2); // version 2 for name & description, if available
     }
+    else if (strcmp(interface, zwp_linux_dmabuf_v1_interface.name) == 0)
+    {
+        dmabuf = (zwp_linux_dmabuf_v1*) wl_registry_bind(registry, name,
+            &zwp_linux_dmabuf_v1_interface, 3);
+    }
+    else if (strcmp(interface, wl_drm_interface.name) == 0)
+    {
+        drm = (wl_drm*) wl_registry_bind(registry, name, &wl_drm_interface, 1);
+        wl_drm_add_listener(drm, &drm_listener, NULL);
+    }
 }
 
 static void handle_global_remove(void*, struct wl_registry *, uint32_t) {
@@ -293,13 +436,11 @@ static uint64_t timespec_to_usec (const timespec& ts)
     return ts.tv_sec * 1000000ll + 1ll * ts.tv_nsec / 1000ll;
 }
 
-static int next_frame(int frame)
-{
-    return (frame + 1) % MAX_BUFFERS;
-}
-
 static InputFormat get_input_format(wf_buffer& buffer)
 {
+    if (use_dmabuf && !use_hwupload) {
+        return INPUT_FORMAT_DMABUF;
+    }
     switch (buffer.format) {
     case WL_SHM_FORMAT_ARGB8888:
     case WL_SHM_FORMAT_XRGB8888:
@@ -319,6 +460,15 @@ static InputFormat get_input_format(wf_buffer& buffer)
     case WL_SHM_FORMAT_ABGR2101010:
     case WL_SHM_FORMAT_XBGR2101010:
         return INPUT_FORMAT_X2BGR10;
+    case WL_SHM_FORMAT_ABGR16161616:
+    case WL_SHM_FORMAT_XBGR16161616:
+        return INPUT_FORMAT_RGBX64;
+    case WL_SHM_FORMAT_ARGB16161616:
+    case WL_SHM_FORMAT_XRGB16161616:
+        return INPUT_FORMAT_BGRX64;
+    case WL_SHM_FORMAT_ABGR16161616F:
+    case WL_SHM_FORMAT_XBGR16161616F:
+        return INPUT_FORMAT_RGBX64F;
     default:
         fprintf(stderr, "Unsupported buffer format %d, exiting.", buffer.format);
         std::exit(0);
@@ -327,13 +477,15 @@ static InputFormat get_input_format(wf_buffer& buffer)
 
 static void write_loop(FrameWriterParams params)
 {
-    /* Ignore SIGINT, main loop is responsible for the exit_main_loop signal */
+    /* Ignore SIGTERM/SIGINT/SIGHUP, main loop is responsible for the exit_main_loop signal */
     sigset_t sigset;
     sigemptyset(&sigset);
-    sigaddset(&sigset, SIGINT);
+    for (auto signo : GRACEFUL_TERMINATION_SIGNALS)
+    {
+        sigaddset(&sigset, signo);
+    }
     pthread_sigmask(SIG_BLOCK, &sigset, NULL);
 
-    int last_encoded_frame = 0;
 #ifdef HAVE_PULSE
     std::unique_ptr<PulseReader> pr;
 #endif
@@ -341,10 +493,14 @@ static void write_loop(FrameWriterParams params)
     while(!exit_main_loop)
     {
         // wait for frame to become available
-        while(buffers[last_encoded_frame].available != true && !exit_main_loop) {
+        while(buffers.encode().ready_encode() != true && !exit_main_loop) {
             std::this_thread::sleep_for(std::chrono::microseconds(1000));
         }
-        auto& buffer = buffers[last_encoded_frame];
+        if (exit_main_loop) {
+            break;
+        }
+
+        auto& buffer = buffers.encode();
 
         frame_writer_pending_mutex.lock();
         frame_writer_mutex.lock();
@@ -354,6 +510,7 @@ static void write_loop(FrameWriterParams params)
         {
             /* This is the first time buffer attributes are available */
             params.format = get_input_format(buffer);
+            params.drm_format = buffer.drm_format;
             params.width = buffer.width;
             params.height = buffer.height;
             params.stride = buffer.stride;
@@ -363,24 +520,44 @@ static void write_loop(FrameWriterParams params)
             if (params.enable_audio)
             {
                 pulseParams.audio_frame_size = frame_writer->get_audio_buffer_size();
+                pulseParams.sample_rate = params.sample_rate;
                 pr = std::unique_ptr<PulseReader> (new PulseReader(pulseParams));
                 pr->start();
             }
 #endif
         }
 
-        bool do_cont = frame_writer->add_frame((unsigned char*)buffer.data,
-            buffer.base_usec, buffer.y_invert);
-        if (!do_cont) {
-            break;
+        bool do_cont = false;
+
+        if (use_dmabuf) {
+            if (use_hwupload) {
+                uint32_t stride = 0;
+                void *map_data = NULL;
+                void *data = gbm_bo_map(buffer.bo, 0, 0, buffer.width, buffer.height,
+                    GBM_BO_TRANSFER_READ, &stride, &map_data);
+                if (!data) {
+                    std::cerr << "Failed to map bo" << std::endl;
+                    break;
+                }
+                do_cont = frame_writer->add_frame((unsigned char*)data,
+                    buffer.base_usec, buffer.y_invert);
+                gbm_bo_unmap(buffer.bo, map_data);
+            } else {
+                do_cont = frame_writer->add_frame(buffer.bo,
+                    buffer.base_usec, buffer.y_invert);
+            }
+        } else {
+            do_cont = frame_writer->add_frame((unsigned char*)buffer.data,
+                buffer.base_usec, buffer.y_invert);
         }
 
         frame_writer_mutex.unlock();
 
-        buffer.available = false;
-        buffer.released = true;
+        if (!do_cont) {
+            break;
+        }
 
-        last_encoded_frame = next_frame(last_encoded_frame);
+        buffers.next_encode();
     }
 
     std::lock_guard<std::mutex> lock(frame_writer_mutex);
@@ -392,7 +569,7 @@ static void write_loop(FrameWriterParams params)
     frame_writer = nullptr;
 }
 
-void handle_sigint(int)
+void handle_graceful_termination(int)
 {
     exit_main_loop = true;
 }
@@ -400,7 +577,7 @@ void handle_sigint(int)
 static bool user_specified_overwrite(std::string filename)
 {
     struct stat buffer;   
-    if (stat (filename.c_str(), &buffer) == 0)
+    if (stat (filename.c_str(), &buffer) == 0 && !S_ISCHR(buffer.st_mode))
     {
         std::string input;
         std::cout << "Output file \"" << filename << "\" exists. Overwrite? Y/n: ";
@@ -432,6 +609,11 @@ static void check_has_protos()
         exit(EXIT_FAILURE);
     }
 
+    if (use_dmabuf && dmabuf == NULL) {
+        fprintf(stderr, "compositor doesn't support linux-dmabuf-unstable-v1\n");
+        exit(EXIT_FAILURE);
+    }
+
     if (available_outputs.empty())
     {
         fprintf(stderr, "no outputs available\n");
@@ -547,21 +729,29 @@ Use Ctrl+C to stop.)");
 #ifdef HAVE_PULSE
     printf(R"(
 
-  -a, --audio [DEVICE]      Starts recording the screen with audio.
-                            [DEVICE] argument is optional.
+  -a, --audio[=DEVICE]      Starts recording the screen with audio.
+                            [=DEVICE] argument is optional.
                             In case you want to specify the pulseaudio device which will capture
                             the audio, you can run this command with the name of that device.
-                            You can find your device by running: pactl list sources | grep Name)");
+                            You can find your device by running: pactl list sources | grep Name
+                            Specify device like this: -a<device> or --audio=<device>)");
 #endif
     printf(R"(
 
-  -c, --codec               Specifies the codec of the video. Supports  GIF output also.
+  -c, --codec               Specifies the codec of the video. These can be found by using:
+                            ffmpeg -encoders
                             To modify codec parameters, use -p <option_name>=<option_value>
-
+  
+  -r, --framerate           Changes framerate to constant framerate with a given value.
+  
   -d, --device              Selects the device to use when encoding the video
                             Some drivers report support for rgb0 data for vaapi input but
                             really only support yuv.
 
+  --no-dmabuf               By default, wf-recorder will try to use only GPU buffers and copies if
+                            using a GPU encoder. However, this can cause issues on some systems. In such
+                            cases, this option will disable the GPU copy and force a CPU one.
+
   -D, --no-damage           By default, wf-recorder will request a new frame from the compositor
                             only when the screen updates. This results in a much smaller output
                             file, which however has a variable refresh rate. When this option is
@@ -573,13 +763,13 @@ Use Ctrl+C to stop.)");
                             while extension .ext . If the extension .ext provided is not
                             recognized by your FFmpeg muxers, the command will fail.
                             You can check the muxers that your FFmpeg installation supports by
-                            running  : ffmpeg -muxers
+                            running: ffmpeg -muxers
 
   -m, --muxer               Set the output format to a specific muxer instead of detecting it
                             from the filename.
 
   -x, --pixel-format        Set the output pixel format. These can be found by running:
-                            *ffmpeg -pix_fmts*
+                            ffmpeg -pix_fmts
 
   -g, --geometry            Selects a specific part of the screen. The format is "x,y WxH".
 
@@ -595,13 +785,26 @@ Use Ctrl+C to stop.)");
                             -p <option_name>=<option_value>
 
   -F, --filter              Specify the ffmpeg filter string to use. For example,
-                            -F hwupload,scale_vaapi=format=nv12 is used for VAAPI.
-
-  -t, --force-yuv           Use the -t or --force-yuv option to force conversion of the data to
-                            yuv format, before sending it to the gpu.
+                            -F scale_vaapi=format=nv12 is used for VAAPI.
 
   -b, --bframes             This option is used to set the maximum number of b-frames to be used.
-                            If b-frames are not supported by your hardware, set this to 0.)" "\n\n" R"(
+                            If b-frames are not supported by your hardware, set this to 0.
+    
+  -B. --buffrate            This option is used to specify the buffers expected framerate. this 
+                            may help when encoders are expecting specific or limited framerate.
+  
+  -C, --audio-codec         Specifies the codec of the audio. These can be found by running:
+                            ffmpeg -encoders
+                            To modify codec parameters, use -P <option_name>=<option_value>
+
+  -X, --sample-format       Set the output audio sample format. These can be found by running: 
+                            ffmpeg -sample_fmts
+  
+  -R, --sample-rate         Changes the audio sample rate in HZ. The default value is 48000.
+  
+  -P, --audio-codec-param   Change the audio codec parameters.
+                            -P <option_name>=<option_value>
+
 Examples:)");
 #ifdef HAVE_PULSE
     printf(R"(
@@ -615,19 +818,19 @@ Examples:)");
                                         current working directory.
 
   - wf-recorder -f <filename>.ext       Records the video. Use Ctrl+C to stop recording.
-                                        The video file will be stored as <outputname>.ext in the
+                                        The video file will be stored as <filename>.ext in the
                                         current working directory.)");
 #ifdef HAVE_PULSE
     printf(R"(
 
   Video and Audio:
 
-  - wf-recorder -a                      Records the audio. Use Ctrl+C to stop recording.
+  - wf-recorder -a                      Records the video and audio. Use Ctrl+C to stop recording.
                                         The video file will be stored as recording.mp4 in the
                                         current working directory.
 
-  - wf-recorder -a -f <filename>.ext    Records the audio. Use Ctrl+C to stop recording.
-                                        The video file will be stored as <outputname>.ext in the
+  - wf-recorder -a -f <filename>.ext    Records the video and audio. Use Ctrl+C to stop recording.
+                                        The video file will be stored as <filename>.ext in the
                                         current working directory.)");
 #endif
     printf(R"(
@@ -664,44 +867,64 @@ void request_next_frame()
     zwlr_screencopy_frame_v1_add_listener(frame, &frame_listener, NULL);
 }
 
+static void parse_codec_opts(std::map<std::string, std::string>& options, const std::string param)
+{
+    size_t pos;
+    pos = param.find("=");
+    if (pos != std::string::npos && pos != param.length() -1)
+    {
+        auto optname = param.substr(0, pos);
+        auto optvalue = param.substr(pos + 1, param.length() - pos - 1);
+        options.insert(std::pair<std::string, std::string>(optname, optvalue));
+    } else
+    {
+        std::cerr << "Invalid codec option " + param << std::endl;
+    }
+}
 
 int main(int argc, char *argv[])
 {
     FrameWriterParams params = FrameWriterParams(exit_main_loop);
-    params.file = "recording.mp4";
+    params.file = "recording." + std::string(DEFAULT_CONTAINER_FORMAT);
     params.codec = DEFAULT_CODEC;
+    params.audio_codec = DEFAULT_AUDIO_CODEC;
+    params.sample_rate = DEFAULT_AUDIO_SAMPLE_RATE;
     params.enable_ffmpeg_debug_output = false;
     params.enable_audio = false;
-    params.force_yuv = false;
     params.bframes = -1;
 
     constexpr const char* default_cmdline_output = "interactive";
     std::string cmdline_output = default_cmdline_output;
+    bool force_no_dmabuf = false;
 
     struct option opts[] = {
-        { "output",          required_argument, NULL, 'o' },
-        { "file",            required_argument, NULL, 'f' },
-        { "muxer",           required_argument, NULL, 'm' },
-        { "pixel-format",    required_argument, NULL, 'x' },
-        { "geometry",        required_argument, NULL, 'g' },
-        { "codec",           required_argument, NULL, 'c' },
-        { "codec-param",     required_argument, NULL, 'p' },
-        { "device",          required_argument, NULL, 'd' },
-        { "filter",          required_argument, NULL, 'F' },
-        { "log",             no_argument,       NULL, 'l' },
-        { "audio",           optional_argument, NULL, 'a' },
-        { "help",            no_argument,       NULL, 'h' },
-        { "force-yuv",       no_argument,       NULL, 't' },
-        { "bframes",         required_argument, NULL, 'b' },
-        { "version",         no_argument,       NULL, 'v' },
-        { "no-damage",       no_argument,       NULL, 'D' },
-        { 0,                 0,                 NULL,  0  }
+        { "output",            required_argument, NULL, 'o' },
+        { "file",              required_argument, NULL, 'f' },
+        { "muxer",             required_argument, NULL, 'm' },
+        { "geometry",          required_argument, NULL, 'g' },
+        { "codec",             required_argument, NULL, 'c' },
+        { "codec-param",       required_argument, NULL, 'p' },
+        { "framerate",         required_argument, NULL, 'r' },
+        { "pixel-format",      required_argument, NULL, 'x' },
+        { "audio-codec",       required_argument, NULL, 'C' },
+        { "audio-codec-param", required_argument, NULL, 'P' },
+        { "sample-rate",       required_argument, NULL, 'R' },
+        { "sample-format",     required_argument, NULL, 'X' },
+        { "device",            required_argument, NULL, 'd' },
+        { "no-dmabuf",         no_argument,       NULL, '&' },
+        { "filter",            required_argument, NULL, 'F' },
+        { "log",               no_argument,       NULL, 'l' },
+        { "audio",             optional_argument, NULL, 'a' },
+        { "help",              no_argument,       NULL, 'h' },
+        { "bframes",           required_argument, NULL, 'b' },
+        { "buffrate",          required_argument, NULL, 'B' },
+        { "version",           no_argument,       NULL, 'v' },
+        { "no-damage",         no_argument,       NULL, 'D' },
+        { 0,                   0,                 NULL,  0  }
     };
 
     int c, i;
-    std::string param;
-    size_t pos;
-    while((c = getopt_long(argc, argv, "o:f:m:x:g:c:p:d:b:la::thvDF:", opts, &i)) != -1)
+    while((c = getopt_long(argc, argv, "o:f:m:g:c:p:r:x:C:P:R:X:d:b:B:la::hvDF:", opts, &i)) != -1)
     {
         switch(c)
         {
@@ -721,10 +944,6 @@ int main(int argc, char *argv[])
                 params.muxer = optarg;
                 break;
 
-            case 'x':
-                params.pix_fmt = optarg;
-                break;
-
             case 'g':
                 selected_region.set_from_string(optarg);
                 break;
@@ -733,6 +952,26 @@ int main(int argc, char *argv[])
                 params.codec = optarg;
                 break;
 
+            case 'r':
+                params.framerate = atoi(optarg);
+                break;
+
+            case 'x':
+                params.pix_fmt = optarg;
+                break;
+
+            case 'C':
+                params.audio_codec = optarg;
+                break;
+
+            case 'R':
+                params.sample_rate = atoi(optarg);
+                break;
+
+            case 'X':
+                params.sample_fmt = optarg;
+                break;
+
             case 'd':
                 params.hw_device = optarg;
                 break;
@@ -741,6 +980,10 @@ int main(int argc, char *argv[])
                 params.bframes = atoi(optarg);
                 break;
 
+            case 'B':
+                params.buffrate = atoi(optarg);
+                break;
+
             case 'l':
                 params.enable_ffmpeg_debug_output = true;
                 break;
@@ -755,26 +998,12 @@ int main(int argc, char *argv[])
 #endif
                 break;
 
-            case 't':
-                params.force_yuv = true;
-                break;
-
             case 'h':
                 help();
                 break;
 
             case 'p':
-                param = optarg;
-                pos = param.find("=");
-                if (pos != std::string::npos && pos != param.length() - 1)
-                {
-                    auto optname = param.substr(0, pos);
-                    auto optvalue = param.substr(pos + 1, param.length() - pos - 1);
-                    params.codec_options[optname] = optvalue;
-                } else
-                {
-                    printf("Invalid codec option %s\n", optarg);
-                }
+                parse_codec_opts(params.codec_options, optarg);
                 break;
 
             case 'v':
@@ -785,6 +1014,14 @@ int main(int argc, char *argv[])
                 use_damage = false;
                 break;
 
+            case 'P':
+                parse_codec_opts(params.audio_codec_options, optarg);
+                break;
+
+            case '&':
+                force_no_dmabuf = true;
+                break;
+
             default:
                 printf("Unsupported command line argument %s\n", optarg);
         }
@@ -806,6 +1043,64 @@ int main(int argc, char *argv[])
     wl_registry_add_listener(registry, &registry_listener, NULL);
     sync_wayland();
 
+    if (params.codec.find("vaapi") != std::string::npos)
+    {
+        std::cout << "using VA-API, trying to enable DMA-BUF capture..." << std::endl;
+
+        // try compositor device if not explicitly set
+        if (params.hw_device.empty())
+        {
+            params.hw_device = drm_device_name;
+        }
+
+        // check we use same device as compositor
+        if (!params.hw_device.empty() && params.hw_device == drm_device_name && !force_no_dmabuf)
+        {
+            use_dmabuf = true;
+        } else if (force_no_dmabuf) {
+            std::cout << "Disabling DMA-BUF as requested on command line" << std::endl;
+        } else {
+            std::cout << "compositor running on different device, disabling DMA-BUF" << std::endl;
+        }
+
+        // region with dmabuf not implemented in wlroots
+        if (selected_region.is_selected())
+        {
+            use_dmabuf = false;
+            std::cout << "region capture not supported with DMA-BUF" << std::endl;
+        }
+
+        if (params.video_filter == "null")
+        {
+            params.video_filter = "scale_vaapi=format=nv12:out_range=full";
+            if (!use_dmabuf)
+            {
+                params.video_filter.insert(0, "hwupload,");
+            }
+        }
+
+        if (use_dmabuf)
+        {
+            std::cout << "enabled DMA-BUF capture, device " << params.hw_device.c_str() << std::endl;
+
+            drm_fd = open(params.hw_device.c_str(), O_RDWR);
+            if (drm_fd < 0)
+            {
+                fprintf(stderr, "failed to open drm device: %m\n");
+                return EXIT_FAILURE;
+            }
+
+            gbm_device = gbm_create_device(drm_fd);
+            if (gbm_device == NULL)
+            {
+                fprintf(stderr, "failed to create gbm device: %m\n");
+                return EXIT_FAILURE;
+            }
+
+            use_hwupload = params.video_filter.find("hwupload") != std::string::npos;
+        }
+    }
+
     check_has_protos();
     load_output_info();
 
@@ -870,23 +1165,18 @@ int main(int argc, char *argv[])
     timespec first_frame;
     first_frame.tv_sec = -1;
 
-    active_buffer = 0;
-    for (auto& buffer : buffers)
-    {
-        buffer.wl_buffer = NULL;
-        buffer.available = false;
-        buffer.released = true;
-    }
-
     bool spawned_thread = false;
     std::thread writer_thread;
 
-    signal(SIGINT, handle_sigint);
+    for (auto signo : GRACEFUL_TERMINATION_SIGNALS)
+    {
+        signal(signo, handle_graceful_termination);
+    }
 
     while(!exit_main_loop)
     {
         // wait for a free buffer
-        while(buffers[active_buffer].released != true) {
+        while(buffers.capture().ready_capture() != true) {
             std::this_thread::sleep_for(std::chrono::microseconds(500));
         }
 
@@ -901,7 +1191,7 @@ int main(int argc, char *argv[])
             break;
         }
 
-        auto& buffer = buffers[active_buffer];
+        auto& buffer = buffers.capture();
         //std::cout << "first buffer at " << timespec_to_usec(get_ct()) / 1.0e6<< std::endl;
 
         if (!spawned_thread)
@@ -919,18 +1209,24 @@ int main(int argc, char *argv[])
         buffer.base_usec = timespec_to_usec(buffer.presented)
             - timespec_to_usec(first_frame);
 
-        buffer.released = false;
-        buffer.available = true;
-
-        active_buffer = next_frame(active_buffer);
+        buffers.next_capture();
     }
 
-    writer_thread.join();
-
-    for (auto& buffer : buffers)
+    if (writer_thread.joinable())
     {
-        if (buffer.wl_buffer)
-            wl_buffer_destroy(buffer.wl_buffer);
+        writer_thread.join();
+    }
+
+    for (size_t i = 0; i < buffers.size(); ++i)
+    {
+        auto buffer = buffers.at(i);
+        if (buffer && buffer->wl_buffer)
+            wl_buffer_destroy(buffer->wl_buffer);
+    }
+
+    if (gbm_device) {
+        gbm_device_destroy(gbm_device);
+        close(drm_fd);
     }
 
     return EXIT_SUCCESS;
